#include <cmath>
#include <iostream>
#include <GL/glew.h>
#include <GL/freeglut.h>
#include <GL/freeglut_ext.h>
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <vector>
#include "simage.h"

#ifndef M_PI
#define M_PI 3.141592
#endif

// VAO
GLuint squareVAO;
GLuint cylinderVAO;
GLuint sphereVAO;
int numCylinderVertices = 0;
int numSphereVertices = 0;

// Shader
GLuint programID;
// filter
int g_filter = 0;

GLuint texture1 = 0;
GLuint textureID = 0;

////////////////////////////////////////////////////////////////////////////////////////////////////

// Create a NULL-terminated string by reading the provided file
static char* readShaderSource(const char* shaderFile)
{
    FILE* fp = fopen(shaderFile, "rb");

    if ( fp == NULL ) { return NULL; }

    fseek(fp, 0L, SEEK_END);
    long size = ftell(fp);

    fseek(fp, 0L, SEEK_SET);
    char* buf = new char[size + 1];
    fread(buf, 1, size, fp);

    buf[size] = '\0';
    fclose(fp);

    return buf;
}

// Create a GLSL program object from vertex and fragment shader files
GLuint InitShader(const char* vShaderFile, const char* fShaderFile)
{
    struct Shader {
	const char*  filename;
	GLenum       type;
	GLchar*      source;
    }  shaders[2] = {
	{ vShaderFile, GL_VERTEX_SHADER, NULL },
	{ fShaderFile, GL_FRAGMENT_SHADER, NULL }
    };

    GLuint program = glCreateProgram();
    
    for ( int i = 0; i < 2; ++i ) {
	Shader& s = shaders[i];
	s.source = readShaderSource( s.filename );
	if ( shaders[i].source == NULL ) {
	    std::cerr << "Failed to read " << s.filename << std::endl;
	    exit( EXIT_FAILURE );
	}

	GLuint shader = glCreateShader( s.type );

	glShaderSource( shader, 1, (const GLchar**) &s.source, NULL );
	glCompileShader( shader );

	GLint  compiled;
	glGetShaderiv( shader, GL_COMPILE_STATUS, &compiled );
	if ( !compiled ) {
	    std::cerr << s.filename << " failed to compile:" << std::endl;
	    GLint  logSize;
	    glGetShaderiv( shader, GL_INFO_LOG_LENGTH, &logSize );
	    char* logMsg = new char[logSize];
	    glGetShaderInfoLog( shader, logSize, NULL, logMsg );
	    std::cerr << logMsg << std::endl;
	    delete [] logMsg;

	    exit( EXIT_FAILURE );
	}

	delete [] s.source;

	glAttachShader( program, shader );
    }

    /* link  and error check */
    glLinkProgram(program);

    GLint  linked;
    glGetProgramiv( program, GL_LINK_STATUS, &linked );
    if ( !linked ) {
	std::cerr << "Shader program failed to link" << std::endl;
	GLint  logSize;
	glGetProgramiv( program, GL_INFO_LOG_LENGTH, &logSize);
	char* logMsg = new char[logSize];
	glGetProgramInfoLog( program, logSize, NULL, logMsg );
	std::cerr << logMsg << std::endl;
	delete [] logMsg;

	exit( EXIT_FAILURE );
    }

    /* use program object */
    glUseProgram(program);

    return program;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
GLuint InitTexture(const char * filename, GLint wrap=GL_REPEAT, GLint magfilter=GL_LINEAR, GLint minfilter=GL_LINEAR)
{
	int imageWidth, imageHeight, numComponents;
	unsigned char * imageData = NULL;
	unsigned char * rescaledImageData = NULL;
	unsigned char * imagePtr = NULL;
	GLsizei xdim2, ydim2;
	GLenum type;

	if (filename)
	{
		imageData = simage_read_image(filename, &imageWidth, &imageHeight, &numComponents);
	}

    xdim2 = 1;
    while (xdim2 <= imageWidth)
        xdim2 *= 2;
    xdim2 /= 2;
    ydim2 = 1;
    while (ydim2 <= imageHeight)
        ydim2 *= 2;
    ydim2 /= 2;
    if ((imageWidth != xdim2) || (imageHeight != ydim2))
        {
        rescaledImageData = simage_resize(imageData, imageWidth, imageHeight, numComponents,
                                           xdim2, ydim2);
        imagePtr = rescaledImageData;
        }
    else
        imagePtr = imageData;

	if (numComponents == 4) type = GL_RGBA;
	else if (numComponents == 3) type = GL_RGB;
	else if (numComponents == 2) type = GL_LUMINANCE_ALPHA;
	else if (numComponents == 1) type = GL_LUMINANCE;
	
	/* init_resources */
	GLuint texture;
	glGenTextures(1, &texture);	// set textureIndex_
	glBindTexture(GL_TEXTURE_2D, texture);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, wrap);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, wrap);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, magfilter);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minfilter);
	if (minfilter == GL_NEAREST || minfilter == GL_LINEAR)
	{
		glTexImage2D(GL_TEXTURE_2D, 0, (GLint) type, xdim2, ydim2,
					 0, type, GL_UNSIGNED_BYTE, imagePtr);
	}
	else 
	{
		//gluBuild2DMipmaps(GL_TEXTURE_2D, numComponents == 3 ? GL_RGB : GL_RGBA, imageWidth, imageHeight,
					//numComponents == 3 ? GL_RGB : GL_RGBA, GL_UNSIGNED_BYTE, imagePtr);
		glTexImage2D(GL_TEXTURE_2D, 0, (GLint) type, xdim2, ydim2,
					 0, type, GL_UNSIGNED_BYTE, imagePtr);
		glGenerateMipmap(GL_TEXTURE_2D); // Generate mipmaps now!!
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, 0);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 4);
	}
	glBindTexture(GL_TEXTURE_2D, 0);

	return texture;
}

////////////////////////////////////////////////////////////////////////////////////////////////////

GLuint SetSquareData()
{
	// Square vertices
	std::vector<glm::vec3> squareVertices;
	// Square normals
	std::vector<glm::vec3> squareNormals;
	// Square textures
	std::vector<glm::vec2> squareTextureCoords;
	// Indices into squareVerts[] buffer for 2 triangles
	std::vector<int> squareIndices;

	// Vertice Positions
	squareVertices.push_back(glm::vec3(-0.75f, -0.75f, 0.0f));
	squareVertices.push_back(glm::vec3(0.75f, -0.75f, 0.0f));
	squareVertices.push_back(glm::vec3(0.75f, 0.75f, 0.0f));
	squareVertices.push_back(glm::vec3(-0.75f, 0.75f, 0.0f));
	// Vertices Normals
	squareNormals.push_back(glm::vec3(0.0f, 0.0f, 1.0f));
	squareNormals.push_back(glm::vec3(0.0f, 0.0f, 1.0f));
	squareNormals.push_back(glm::vec3(0.0f, 0.0f, 1.0f));
	squareNormals.push_back(glm::vec3(0.0f, 0.0f, 1.0f));
	// Vertices Textures
	squareTextureCoords.push_back(glm::vec2(0.0f, 0.0f));
	squareTextureCoords.push_back(glm::vec2(1.0f, 0.0f));
	squareTextureCoords.push_back(glm::vec2(1.0f, 1.0f));
	squareTextureCoords.push_back(glm::vec2(0.0f, 1.0f));
	// Vertice Indices
	squareIndices.push_back(0);
	squareIndices.push_back(1);
	squareIndices.push_back(2);
	squareIndices.push_back(0);
	squareIndices.push_back(2);
	squareIndices.push_back(3);

	// VAO
	GLuint vao;
	glGenVertexArrays(1, &vao);
	glBindVertexArray(vao);

	// VBO
	GLuint vbo[3];
	glGenBuffers(3, &vbo[0]);
	glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
	glBufferData(GL_ARRAY_BUFFER, 4*sizeof(glm::vec3), &squareVertices[0], GL_STATIC_DRAW); // vertex position
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, 0);
	glEnableVertexAttribArray(0);
	glBindBuffer(GL_ARRAY_BUFFER, vbo[1]);
	glBufferData(GL_ARRAY_BUFFER, 4*sizeof(glm::vec3), &squareNormals[0], GL_STATIC_DRAW); // vertex normal
	glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, 0);
	glEnableVertexAttribArray(1);
	glBindBuffer(GL_ARRAY_BUFFER, vbo[2]);
	glBufferData(GL_ARRAY_BUFFER, 4*sizeof(glm::vec2), &squareTextureCoords[0], GL_STATIC_DRAW); // vertex texture coordinates
	glVertexAttribPointer(2, 2, GL_FLOAT, GL_FALSE, 0, 0);
	glEnableVertexAttribArray(2);

	// IBO
	GLuint ibo;
	glGenBuffers(1, &ibo);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ibo);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, 6*sizeof(int), &squareIndices[0], GL_STATIC_DRAW);

	squareVertices.clear();
	squareNormals.clear();
	squareTextureCoords.clear();

	return vao;
}

GLuint SetCylinderData(float radius=1.0f, float height=2.0f, int slices=16)
{
	std::vector<glm::vec3> cylinderVertices;
	std::vector<glm::vec3> cylinderNormals;
	std::vector<glm::vec2> cylinderTextureCoords;

	glm::vec3 vertex;
	float xTexCoord = 0.0f;

	float theta = (float) (2*M_PI/slices);
	for (int i = 0; i <= slices; i++) 
	{
		vertex[0] = radius * cosf(theta * i);
		vertex[1] = -height/2.0f;
		vertex[2] = radius * sinf(theta * i);
		cylinderVertices.push_back(vertex);
		vertex[1] = 0.0f;
		vertex = glm::normalize(vertex);
		cylinderNormals.push_back(vertex);
		cylinderTextureCoords.push_back(glm::vec2(xTexCoord, 0.0f));
		numCylinderVertices++;

		vertex[0] = radius * cosf(theta * i);
		vertex[1] = height/2.0f;
		vertex[2] = radius * sinf(theta * i);
		cylinderVertices.push_back(vertex);
		vertex[1] = 0.0f;
		vertex = glm::normalize(vertex);
		cylinderNormals.push_back(vertex);
		cylinderTextureCoords.push_back(glm::vec2(xTexCoord, 1.0f));
		numCylinderVertices++;

		xTexCoord += 1.0f/slices;
	}

	// VAO & VBOs
	GLuint vao;
	glGenVertexArrays(1, &vao); 
	glBindVertexArray(vao);

	GLuint vbo[3];
	glGenBuffers(3, &vbo[0]);
	glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
	glBufferData(GL_ARRAY_BUFFER, numCylinderVertices*sizeof(glm::vec3), &cylinderVertices[0], GL_STATIC_DRAW); // vertex position
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, 0);
	glEnableVertexAttribArray(0);
	glBindBuffer(GL_ARRAY_BUFFER, vbo[1]);
	glBufferData(GL_ARRAY_BUFFER, numCylinderVertices*sizeof(glm::vec3), &cylinderNormals[0], GL_STATIC_DRAW); // vertex position
	glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, 0);
	glEnableVertexAttribArray(1);
	glBindBuffer(GL_ARRAY_BUFFER, vbo[2]);
	glBufferData(GL_ARRAY_BUFFER, numCylinderVertices*sizeof(glm::vec2), &cylinderTextureCoords[0], GL_STATIC_DRAW); // vertex position
	glVertexAttribPointer(2, 2, GL_FLOAT, GL_FALSE, 0, 0);
	glEnableVertexAttribArray(2);

	glBindVertexArray(0);
	glBindVertexArray(1);
	glBindVertexArray(2);
	cylinderVertices.clear();
	cylinderNormals.clear();
	cylinderTextureCoords.clear();

	return vao;
}

GLuint SetSphereData(float radius=0.75f, int slices=16, int stacks=16)
{
	std::vector<glm::vec3> sphereVertices;
	std::vector<glm::vec3> sphereNormals;
	std::vector<glm::vec2> sphereTextureCoords;

	glm::vec3 vertex(0.0f, 0.0f, 0.0f);
	float lonstep = (float) (M_PI/stacks);
	float latstep = (float) (M_PI/slices);
	float lon, lat;
	glm::vec2 texCoord(0.0f, 0.0f);
	float dstack = 1.0f/stacks;
	float dslice = 1.0f/slices;
	float texX, texY;

	for (lon = 0.0f, texX = 0.0f; lon <= 2*M_PI; lon += (lonstep), texX += (dstack))
	{
		for (lat = 0.0, texY = 0.0f; lat <= M_PI + latstep; lat += (latstep), texY += (dslice))
		{
			vertex[0] = radius * cosf(lon)*sinf(lat);
			vertex[1] = radius * sinf(lon)*sinf(lat);
			vertex[2] = radius * cosf(lat);
			sphereVertices.push_back(vertex); // vertex position
			vertex = glm::normalize(vertex);
			sphereNormals.push_back(vertex); // vertex normal
			texCoord[0] = texX;
			texCoord[1] = texY;
			sphereTextureCoords.push_back(texCoord); // vertex texture coordnate
			numSphereVertices++;

			vertex[0] = radius * cosf(lon + lonstep)*sinf(lat);
			vertex[1] = radius * sinf(lon + lonstep)*sinf(lat);
			vertex[2] = radius * cosf(lat);
			sphereVertices.push_back(vertex); // vertex position
			vertex = glm::normalize(vertex);
			sphereNormals.push_back(vertex); // vertex normal
			texCoord[0] = texX + dstack;
			texCoord[1] = texY;
			sphereTextureCoords.push_back(texCoord); // vertex texture coordnate
			numSphereVertices++;
		}
	}

	// VAO & VBOs
	GLuint vao;
	glGenVertexArrays(1, &vao); 
	glBindVertexArray(vao);

	GLuint vbo[3];
	glGenBuffers(3, &vbo[0]);
	glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
	glBufferData(GL_ARRAY_BUFFER, numSphereVertices*sizeof(glm::vec3), &sphereVertices[0], GL_STATIC_DRAW); // vertex position
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, 0);
	glEnableVertexAttribArray(0);
	glBindBuffer(GL_ARRAY_BUFFER, vbo[1]);
	glBufferData(GL_ARRAY_BUFFER, numSphereVertices*sizeof(glm::vec3), &sphereNormals[0], GL_STATIC_DRAW); // vertex position
	glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, 0);
	glEnableVertexAttribArray(1);
	glBindBuffer(GL_ARRAY_BUFFER, vbo[2]);
	glBufferData(GL_ARRAY_BUFFER, numSphereVertices*sizeof(glm::vec2), &sphereTextureCoords[0], GL_STATIC_DRAW); // vertex position
	glVertexAttribPointer(2, 2, GL_FLOAT, GL_FALSE, 0, 0);
	glEnableVertexAttribArray(2);

	glBindVertexArray(0);
	glBindVertexArray(1);
	glBindVertexArray(2);
	sphereVertices.clear();
	sphereNormals.clear();
	sphereTextureCoords.clear();

	return vao;
}

void init( void )
{
    // Load shaders and use the resulting shader program
    programID = InitShader( "StandardShading.vert", "StandardShading3.frag" );
	glUseProgram(programID);

	// Set data
	texture1 = InitTexture("opengl.jpg");

	squareVAO = SetSquareData();
	cylinderVAO = SetCylinderData();
	sphereVAO = SetSphereData();

	glClearColor( 0.5, 0.7, 0.9, 1.0 ); //  background
}

void drawAlphaTexSquare(int texture)
{
	// Bind our texture in Texture Unit 0
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, texture);
	// Enable blending
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	// Draw square
	glBindVertexArray(squareVAO);
	glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, 0);       // draw filled
	glBindVertexArray(0);
	glDisable(GL_BLEND);
	glBindTexture(GL_TEXTURE_2D, 0);
}

void drawTexSquare(int texture)
{
	// Bind our texture in Texture Unit 0
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, texture);
	// Draw square
	glBindVertexArray(squareVAO);
	glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, 0);       // draw filled
	glBindTexture(GL_TEXTURE_2D, 0);
}

void drawTexCylinder(int texture)
{
	// Bind our texture in Texture Unit 0
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, texture);
	// Draw square
	glBindVertexArray(cylinderVAO);
	glDrawArrays(GL_TRIANGLE_STRIP, 0, numCylinderVertices);	// draw cylinder
	glBindTexture(GL_TEXTURE_2D, 0);
}

void drawTexSphere(int texture)
{
	// Bind our texture in Texture Unit 0
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, texture);
	// Draw square
	glBindVertexArray(sphereVAO);
	glDrawArrays(GL_TRIANGLE_STRIP, 0, numSphereVertices);	// draw sphere
	glBindTexture(GL_TEXTURE_2D, 0);
}

void display( void )
{
    glClear( GL_COLOR_BUFFER_BIT );     // clear the window

	// Projection matrix : 45?Field of View, 4:3 ratio, display range : 0.1 unit <-> 100 units
	glm::mat4 Projection = glm::perspective(45.0f, 1.0f, 0.1f, 100.0f);
	// Camera matrix
	glm::mat4 View       = glm::lookAt(
								glm::vec3(0,0,4), // Camera is at (4,3,3), in World Space
								glm::vec3(0,0,0), // and looks at the origin
								glm::vec3(0,1,0)  // Head is up (set to 0,-1,0 to look upside-down)
						   );
	// Model matrix : an identity matrix (model will be at the origin)
	glm::mat4 Model      = glm::mat4(1.0f);
	// Our ModelViewProjection : multiplication of our 3 matrices
	glm::mat4 MVP        = Projection * View * Model; // Remember, matrix multiplication is the other way around


    glUseProgram(programID);
	// Get a handle for our "gMVP" uniform
	GLuint projectionMatrixID = glGetUniformLocation(programID, "gProjection");
	glUniformMatrix4fv(projectionMatrixID, 1, GL_FALSE, &Projection[0][0]);
	// Get a handle for our "gView" uniform
	GLuint viewMatrixID = glGetUniformLocation(programID, "gView");
	glUniformMatrix4fv(viewMatrixID, 1, GL_FALSE, &View[0][0]);
	// Get a handle for our "gModel" uniform
	GLuint modelMatrixID = glGetUniformLocation(programID, "gModel");
	glUniformMatrix4fv(modelMatrixID, 1, GL_FALSE, &Model[0][0]);

	glm::vec4 lightPos = glm::vec4(4,4,4,1);
	// Get a handle for our "gLightPosition" uniform
	GLuint lightID = glGetUniformLocation(programID, "gLightPosition");
	glUniform4f(lightID, lightPos.x, lightPos.y, lightPos.z, lightPos.w);

	// Get a handle for our "gTextureSampler" uniform
	textureID  = glGetUniformLocation(programID, "gTextureSampler");
	glUniform1i(textureID, 0);

	// Get a handle for our "gMaterialDiffuseColor" uniform
	GLuint materialID = glGetUniformLocation(programID, "gMaterialDiffuseColor");

	// tex cylinder
	Model = glm::translate(glm::mat4(1.0f),glm::vec3(0.0f, 0.0f, -1.0f));
	glUniformMatrix4fv(modelMatrixID, 1, GL_FALSE, &Model[0][0]);
	glUniform4f(materialID, 1.0f, 1.0f, 1.0f, 1.0f);	
	drawTexCylinder(texture1);

	// tex sphere
	Model = glm::translate(glm::mat4(1.0f),glm::vec3(0.75f, 0.0f, 0.0f));
	glUniformMatrix4fv(modelMatrixID, 1, GL_FALSE, &Model[0][0]);
	glUniform4f(materialID, 1.0f, 1.0f, 1.0f, 1.0f);	
	drawTexSphere(texture1);

	// blended tex square
	glEnable(GL_CULL_FACE);
	glEnable(GL_BLEND);
	if (g_filter == 0)                     // Default (no blending) = Cs*1
		glBlendFunc(GL_ONE, GL_ZERO);
	else if (g_filter == 1)                // Draw background only = Cd*1
		glBlendFunc(GL_ZERO, GL_ONE);
	else if (g_filter == 2)                // Cs*1 + Cd*1 
		glBlendFunc(GL_ONE, GL_ONE);
	else if (g_filter == 3)                // Alpha blending (back-to-front) = Cs*As + Cd*(1-As)
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	else if (g_filter == 4)                // Additive blending (brighten the scene) = Cs*As + Cd*1
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	else if (g_filter == 5)                // Multiplicative blending = Cd*Cs
		glBlendFunc(GL_ZERO, GL_SRC_COLOR);
	else if (g_filter == 6)                // darken the scene = Cd*As
		glBlendFunc(GL_ZERO, GL_SRC_ALPHA);
	else if (g_filter == 7)                 // Invert all the colors = Cs*(1-Cd)
		glBlendFunc(GL_ONE_MINUS_DST_COLOR, GL_ZERO);
	Model = glm::translate(glm::mat4(1.0f),glm::vec3(-0.5f, 0.0f, 1.0f));
	glUniformMatrix4fv(modelMatrixID, 1, GL_FALSE, &Model[0][0]);
	glUniform4f(materialID, 1.0f, 1.0f, 1.0f, 0.5f);	
	drawTexSquare(texture1);
	glDisable(GL_BLEND);

	glutSwapBuffers();
}

void keyboard( unsigned char key, int x, int y )
{
	switch (key)  
	{    
	case 0x1B:    
	case 'q':    
	case 'Q':      
		exit(0);     
	case ' ':
		g_filter = (g_filter + 1) % 8;
		printf("filter=%d\n", g_filter);
		break;  
	}
	glutPostRedisplay();
}



int main( int argc, char **argv )
{
	glutInit(&argc, argv);
	glutInitDisplayMode( GLUT_RGBA | GLUT_DOUBLE );
    glutInitWindowSize( 512, 512 );
    glutCreateWindow(argv[0]);
    
	glewInit();
    init();

    glutDisplayFunc( display );
    glutKeyboardFunc( keyboard );


    glutMainLoop();
    return 0;
}
